import os, sys, sanic, socket

from os import environ as env

from IzzyLib import logging
from IzzyLib.misc import DefaultDict, DotDict, Path, Boolean, GetIp, RandomGen
from IzzyLib.template import Template
from envbash import load_envbash
from jinja2 import Environment
from hamlish_jinja import HamlishExtension

from . import __version__ as VERSION


dbtypes = ['sqlite', 'postgresql', 'mysql', 'mssql']


if getattr(sys, 'frozen', False):
	script_path = Path(sys.executable).parent()

else:
	script_path = Path(__file__).parent()

store_path = script_path.parent(True).join('data')
pyenv = env.get('PYENV', 'dev').lower()
pyenv_path = store_path.join(f'{pyenv}.env')

try:
	if pyenv not in ['prod', 'dev']:
		logging.error('Invalid "PYENV" value. Please use "prod" or "dev"')
		sys.exit()

	load_envbash(pyenv_path.str())


except FileNotFoundError:
	logging.warning('Bash environment file not found:', pyenv_path.str())


listen = env.get('LISTEN', GetIp())
port = int(env.get('PORT', 8020))
web_domain = env.get('WEB_DOMAIN', env.get('DOMAIN', f'{listen}:{port}'))
connections = int(env.get('DB_CONNECTIONS', 25))

config = DotDict({
	'env': pyenv,
	'path': script_path,
	'data': store_path,
	'frontend': script_path.join('frontend', True),
	'version': VERSION,
	'dbversion': 20210103,
	'listen': listen,
	'port': port,
	'web_domain': web_domain,
	'domain': env.get('DOMAIN', web_domain),
	'salt': env.get('PASS_SALT', RandomGen(40)),
	'secret': env.get('FORWARDED_SECRET', RandomGen(20)),
	'agent': f'sanic/{sanic.__version__} (Barkshark-Social/{VERSION}; +https://{web_domain}/)',
	'dbtype': env.get('DB_TYPE', 'sqlite'),
	'log': DotDict({
		'level': env.get('LOG_LEVEL', 'INFO').upper(),
		'errors': Boolean(env.get('LOG_ERRORS')),
		'date': Boolean(env.get('LOG_DATE', True))
	}),
	'vars': DotDict({
		'max_chars': env.get('MAX_CHARS', 69420),
		'posts': env.get('PROFILE_POSTS', 20)
	}),
	'proxy': DotDict({
		'enabled': Boolean(env.get('PROXY_ENABLED', False)),
		'ptype': env.get('PROXY_TYPE', 'https'),
		'host': env.get('PROXY_HOST', None),
		'port': env.get('PROXY_PORT', 443)
	}),
	'rd': DotDict({
		'host': env.get('REDIS_HOST', 'localhost'),
		'port': int(env.get('REDIS_PORT', 6379)),
		'user': env.get('REDIS_USER', env.get('USER', 'social')),
		'password': env.get('REDIS_PASSWORD'),
		'database': int(env.get('REDIS_DATABASE', 0)),
		'prefix': env.get('REDIS_PREFIX', 'social'),
		'maxconnections': int(env.get('REDIS_CONNECTIONS', 25))
	}),
	'db': DotDict({
		'host': env.get('DB_HOST', '/var/run/postgresql'),
		'port': int(env.get('DB_PORT', 5432)),
		'user': env.get('DB_USER', env.get('USER', 'social')),
		'password': env.get('DB_PASS'),
		'database': env.get('DB_DATABASE', 'social'),
		'maxconnections': int(env.get('DB_CONNECTIONS', 25))
	}),
	'sqfile': store_path.join(env.get('SQ_DATABASE', 'database.sqlite3'))
})


if config.dbtype not in dbtypes:
	dbtypes_str = ', '.join(dbtypes)
	logging.error(f'Invalid dbtype : {config.dbtype}')
	os.exit()

# Delete me later
config.update({
	'log_level': config.log.level,
	'log_errors': config.log.errors,
	'log_date': config.log.date
})


if not config['log_date']:
	log_date = ''

else:
	log_date = '[%(asctime)s] '

logging.setConfig({'level': config.log.level})
config.data.join('media', True).mkdir()
config.data.join('template', True).mkdir()

config.template = Template(
	search=[
		config.data.join('template', True),
		config.frontend.str()
	],
	autoescape=False
)

def Context(context, globals):
	try:
		user = context.request.ctx.cookie_user
	except AttributeError:
		user = None

	with globals.db.session() as s:
		settings = s.get.configs(cached=False)

	return {
		'cookie_user': user,
		'menu_left': user.config.get('menu_left', False) if user else False,
		'color_theme': context.request.cookies.get('theme', 'pink'),
		'settings': DefaultDict(settings),
		'var': {'base': 'https://'+config.web_domain}
	}


config.template.func_context = Context
