import sqlite3, sys, traceback

from IzzyLib import logging
from IzzyLib.misc import DotDict
from contextlib import contextmanager
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from .schema import table
from .migrate import upgrade

from . import functions
from ..config import config


class Row(DotDict):
	_filter_columns = lambda self, row: [attr for attr in dir(row) if not attr.startswith('_') and attr != 'metadata']


	def __init__(self, row, db):
		if not row:
			return

		self._db = db
		self._table_name = row.__tablename__
		self._table = db.table[row.__tablename__]
		self._columns = self._filter_columns(row)

		super().__init__()

		for k in self._columns:
			v = getattr(row, k)

			if type(v) == dict:
				v = DotDict(v)

			self[k] = v

		self.__run__(row, db)


	def __run__(self, row, db):
		pass


	def _filter_data(self):
		data = {k: v for k,v in self.items() if k in self._columns}

		for k,v in self.items():
			if v.__class__ == DotDict:
				data[k] = v.asDict()

		return data


	def _update(self, new_data={}, **kwargs):
		kwargs.update(new_data)

		for k,v in kwargs.items():
			self[k] = v


	def delete(self):
		with self._db.session() as s:
			row = s.query(self._table).filter_by(id=self.id)
			row.delete()

			return True if not row.all() else False


	def delete_session(self, s):
		row = s.query(self._table).filter_by(id=self.id)
		row.delete()

		return True if not row.all() else False


	def update(self, dict_data={}, **data):
		self._update(data)
		self._update(dict(dict_data))

		data = self._filter_data()

		with self._db.session() as s:
			s.query(self._table).filter_by(id=self.id).update(data)


	def update_session(self, s, dict_data={}, **data):
		self._update(data)
		self._update(dict_data)

		data = self._filter_data()

		row = s.query(self._table).filter_by(id=self.id)
		row.update(data)


class User(Row):
	def __run__(self, row, db):
		domain = db.fetch('domain', id=self.domainid)

		self.domain = domain.domain
		self.avatar = 'static/img/one.png'
		self.header = 'static/img/one.png'

		avatar = row.config.get('avatar')
		header = row.config.get('header')

		avatar_path = f'media/avatar/{self.domain}/{self.handle[0]}/{avatar}.png'
		header_path = f'media/header/{self.domain}/{self.handle[0]}/{header}.png'

		if config.data.join(avatar_path).exists():
			self.avatar = avatar_path

		if config.data.join(header_path).exists():
			self.header = header_path


class Status(Row):
	def __run__(self, row, db):
		self.user = db.fetch('user', id=row.userid)

		if not self.user:
			logging.warning('db.Status: Could not find user from post')


class Session(object):
	def __init__(self, db):
		self.db = db
		self.cache = self.db.cache
		self.classes = self.db.classes
		self.session = sessionmaker(bind=db.db)()
		self.table = self.db.table

		# session aliases
		self.s = self.session
		self.commit = self.s.commit
		self.rollback = self.s.rollback
		self.query = self.s.query
		self.execute = self.s.execute

		# extra function aliases
		self.get = functions.Get(self)
		self.put = functions.Put(self)


	def __enter__(self):
		return self


	def __exit__(self, exctype, value, tb):
		if tb:
			self.rollback()
			#print(exctype, value)
			#traceback.print_tb(tb)

		else:
			self.commit()


	def count(self, table_name, **kwargs):
		return self.s.query(self.table[table_name]).filter_by(**kwargs).count()


	def execute(self, string, values=[]):
		return self.s.execute(string, values)


	def fetch(self, table_name, single=True, **kwargs):
		RowClass = self.classes.get(table_name.capitalize(), Row)

		rows = self.s.query(self.table[table_name]).filter_by(**kwargs).all()

		if single:
			return RowClass(rows[0], db) if len(rows) > 0 else None

		return [RowClass(row, db) for row in rows]


	def insert(self, table_name, **kwargs):
		table = self.table[table_name]

		if getattr(table, 'timestamp', None) and not kwargs.get('timestamp'):
			kwargs['timestamp'] = datetime.now()

		data = self.s.add(table(**kwargs))
		return self.fetch(table_name, **kwargs)


	def DropTables(self):
		tables = self.GetTables()

		for table in tables:
			self.s.execute(f'DROP TABLE {table}')


	def GetTables(self):
		rows = self.s.execute("SELECT name FROM sqlite_master WHERE type IN ('table','view') and name NOT LIKE 'sqlite_%'")
		return [row[0] for row in rows]


class DataBase(object):
	Row = Row
	cache = DotDict()
	classes = DotDict()
	classes.User = User
	classes.Status = Status


	def __init__(self, sqfile=None, tables=None):
		self.db = create_engine(sqfile)
		self.table = tables
		self.session = lambda : Session(self)

		self.get = lambda : self.session().get
		self.put = lambda : self.session().put

		for table in tables:
			self.cache[table] = DotDict()


	def execute(self, *args, **kwargs):
		with self.session() as s:
			return s.execute(*args, **kwargs)


	def fetch(self, *args, **kwargs):
		with self.session() as s:
			return s.fetch(*args, **kwargs)


	def count(self, *args, **kwargs):
		with self.session() as s:
			return s.count(*args, **kwargs)


	def GetTables(self, *args, **kwargs):
		with self.session() as s:
			return s.GetTables(*args, **kwargs)


	def DropTables(self, *args, **kwargs):
		with self.session() as s:
			return s.DropTables(*args, **kwargs)


	@contextmanager
	def simple_session(self):
		session = sessionmaker(bind=self.db)()

		try:
			yield session

		except Exception:
			traceback.print_exc()
			session.rollback()

		finally:
			session.commit()


db = DataBase(f'sqlite:///{config.sqfile}', table)
with db.session() as s:
	config_version = s.get.version()

if config_version in [0, None]:
	pass

elif config_version < config.dbversion:
	logging.error('Run "python3 -m social.manage migrate" to upgrade the database schema')
	sys.exit()

else:
	with db.session() as s:
		# not sure how many users is too many for sqlite, but 5 is still kinda low
		# doesn't count system accounts btw
		if config.dbtype == 'sqlite' and s.get.stats().user_count >= 5:
			logging.warning('The sqlite backend is in use with 5 or more users which might be unstable')
			logging.warning('It is highly recommended to convert the database to postgresql')

		for k,v in s.get.configs(False).items():
			db.cache.config[k] = v

		config.domainid = s.get.domainid(config.domain)
