from IzzyLib import logging
from IzzyLib.misc import Boolean, DefaultDict, DotDict, RandomGen
from datetime import datetime, timedelta
from sqlalchemy import func
from sqlalchemy_paginator import Paginator

from ..config import config
from ..functions import Client, FetchActor, FetchApi, GenKey, mkhash


subtypes = {
	'str': str,
	'int': int,
	'bool': Boolean,
	'datetime': int
}


config_defaults = {
	'admin': None,
	'name': 'Barkshark Social',
	'description': 'Barkshark Social is an upcoming Activity Pub server with a focus on fredom of association, low resource usage, and cusomizability.',
	'description_short': None,
	'char_limit': 5000,
	'bio_limit': 1000,
	'table_limit': 8,
	'emoji_size_limit': 100,
	'secure': True,
	'require_approval': True,
	'domain_approval': False,
	'open': True,
	'robotstxt': 'User-agent: *\nDisallow: /'
}

user_config_defaults = {
	'public': False,
	'locked': False,
	'private': False,
	'info_table': {},
	'avatar': None,
	'header': None
}


def SetupFuncs(self, session):
	self.session = session
	self.s = session.s
	self.classes = session.db.classes
	self.db = session.db

	self.cache = session.cache
	self.count = session.count
	self.fetch = session.fetch
	self.insert = session.insert
	self.table = session.table

	self.query = session.s.query
	self.execute = session.s.execute


class Get():
	def __init__(self, session):
		SetupFuncs(self, session)


	def __parse_config(self, row):
		typefunc = subtypes[row.type] if row.type else str
		value = typefunc(row.value)

		if row.key == 'emoji_size_limit':
			value = value * 1024

		return datetime.fromtimestamp(value) if row.type == 'datetime' else value


	def config(self, key, default=None):
		cached = self.cache.config.get(key)

		if cached:
			return cached

		row = self.fetch('config', key=key)

		if row:
			var = self.__parse_config(row)
			self.cache.config[key] = var
			return var

		if not default:
			value = config_defaults.get(key)
			return value * 1000 if key == 'emoji_size_limit' else value

		return default


	def configs(self, cached=True):
		data = DotDict(config_defaults.copy())

		if cached:
			data.update(self.cache.config)
			return DotDict(data)

		rows = self.fetch('config', single=False)

		for row in rows:
			data[row.key] = self.__parse_config(row)

		return data


	def domain(self, domain=None):
		web_domain = domain if domain else config.web_domain
		domain = domain if domain else config.domain
		row = self.fetch('domain', web_domain=web_domain)

		if not row:
			row = self.fetch('domain', domain=domain)

		return row


	def domainid(self, domain=None):
		row = self.domain(domain)
		return row.id if row else None


	def emoji(self, shortcode=None, domain=None):
		domain = domain if domain else config.domain
		domainid = self.domainid(domain)
		cachekey = domain+shortcode
		cache = self.cache.emoji.get(cachekey)

		if cache:
			return cache

		if not domainid:
			logging.error('get.emoji: Cannot find domain:', domain)
			return

		row = self.fetch('emoji', shortcode=shortcode, domainid=domainid)

		if row:
			row.path = f'media/emojis/{domainid}/{row.shortcode[0]}/{row.shortcode}.png'
			self.cache.emoji[cachekey] = row

			return self.cache.emoji[cachekey]

		return


	def user_count(self, domain=None):
		domain = self.fetch('domain', domain=config.domain if not domain else domain)
		return self.query(self.table.user).filter_by(domainid=domain.id).count()


	def domain_count(self):
		return self.query(self.table.domain).count()


	def stats(self):
		domainid = self.domainid(config.domain)
		sys_accts = [row.id for row in self.fetch('user', permissions=0, single=False)]
		user = self.table.user
		status = self.table.status

		today = datetime.today()

		users = self.query(user).filter(user.domainid == domainid, user.permissions > 0)
		statuses = self.query(status).filter(status.domainid == domainid, status.userid.notin_(sys_accts))

		data = DefaultDict()
		data.user_count = users.count()
		data.users_month = users.filter(user.timestamp >= (datetime.today() - timedelta(days=30))).count()
		data.users_halfyear = users.filter(user.timestamp >= (datetime.today() - timedelta(days=30*6))).count()
		data.status_count = statuses.count()

		return data


	def status_count(self, domain=None):
		domain = self.fetch('domain', domain=domain)
		return self.query(self.table.status).filter_by(domainid=domain.id).count()


	def statuses(self, handle=None, user=None, domain=None, count=20, desc=True):
		if not any([handle, user]):
			logging.error('db.get.statuses: Please specify a user or handle')
			return

		if handle and user:
			logging.error('db.get.statuses: Only specify a handle OR user')
			return

		if handle:
			domainid = self.domainid(domain)
			user = self.fetch('user', handle=handle, domainid=domainid)

		if not user and not handle:
			logging.debug('db.get.statuses: Cannot find user')
			return

		if not user and handle:
			logging.debug('db.get.status: Cannot find user:', handle)
			return

		table = self.table.status
		rows = self.query(table).filter(table.userid == user.id)

		if desc:
			rows = rows.order_by(table.id.desc())

		return Paginator(rows, count)


	def token(self, token_header, table='token'):
		response = {'user': None, 'token': None}
		response['token'] = self.fetch(table, token=token_header)

		if response['token']:
			response['user'] = self.fetch('user', id=response['token'].userid)

		if not response['user'] and response['token']:
			response['token'].delete_session(self.s)
			response['token'] = None

		return response


	def user(self, handle, domain=None, single=True):
		domainid = self.domainid(domain)

		if not domainid:
			return

		row = self.fetch('user', lower=handle.lower(), domainid=domainid)
		row = self.query(self.table.user).filter(func.lower(self.table.user.handle) == func.lower(handle) and self.table.user.domainid == domainid)
		return self.classes.User(row.one_or_none(), self.db)


	def version(self):
		try:
			rows = self.execute("SELECT name FROM sqlite_master WHERE name NOT LIKE 'sqlite_%'").fetchall()
			tables = [row[0] for row in rows]

			if 'config' not in tables:
				return 0

			row = self.config('version')

			if not row:
				return 0

			return row

		except IndexError:
			return 0


class Put():
	def __init__(self, session):
		SetupFuncs(self, session)


	def app(self, redirect_uri, scope, name, url, client_id=RandomGen(40, '-_'), client_secret=RandomGen(40, '-_')):
		data = {
			'client_id': client_id,
			'client_secret': client_secret,
			'redirect_uri': redirect_uri,
			'scope': scope,
			'name': name,
			'url': url,
			'timestamp': datetime.now()
		}

		row = self.fetch('app', client_id=client_id, client_secret=client_secret)

		if row:
			row.update_session(s, data)
			return row

		self.insert('app', **data)
		return self.fetch('app', client_id=data['client_id'])


	def config(self, key, value, subtype='str'):
		row = self.fetch('config', key=key)
		typefunc = subtypes[row.type] if row else str

		if subtype == 'bool':
			new_value = 'True' if Boolean(value) else 'False'

		else:
			new_value = int(datetime.timestamp(value)) if subtype == 'datetime' else typefunc(value)

		if row:
			if row.value == new_value:
				return

			row.value = new_value
			row.update_session(self.s)

		else:
			self.insert('config',
				key=key,
				value=new_value,
				type=subtype
			)

		self.cache.config[key] = value


	def configs(self, *args):
		for config in args:
			key = config.get('key')
			value = config.get('value')
			subtype = config.get('subtype')

			if None in [key, value]:
				logging.warning(f'db.put.configs: Missing key or value for config: key={key}, value={value}')
				continue

			self.config(key, value, subtype)


	def cookie(self, userid, address=None, agent=None, access=datetime.now(), token=RandomGen()):
		row = self.fetch('cookie', token=token, userid=userid)

		if row:
			row.update(access=access)

		else:
			self.insert('cookie',
				token = token,
				userid = userid,
				address = address,
				agent = agent,
				access = access
			)

			row = self.fetch('cookie', token=token, userid=userid)

		return row


	def domain(self, domain, web_domain):
		api = FetchApi(web_domain)
		description = api.description if api.description else api.short_description
		admin = None

		if api.contact_account:
			admin = api.contact_account['acct']

		self.insert('domain',
			domain = domain,
			web_domain = web_domain,
			name = api.title,
			admin = admin,
			description = description,
			short_description = api.short_description,
			timestamp = datetime.now()
		)


	def emoji(self, filename, shortcode, domain=None, display=None, enabled=None):
		domain = domain if domain else config.domain
		domainrow = self.fetch('domain', domain=domain)
		cachekey = domain+shortcode

		if not domainrow:
			logging.error(f'session.put.emoji: Not a known domain: {domain}')

		row = self.query('emoji', shortcode=shortcode, domainid=domainrow.id)

		if row:
			new_data = {}

			if display != None:
				new_data['display'] = display
			if enabled != None:
				new_data['enabled'] = enabled

			if new_data:
				row.update(new_data)

		else:
			self.insert('emoji',
				shortcode = shortcode,
				domainid = domainrow.id,
				display = display,
				enabled = enabled,
				filename = filename,
				timestamp = datetime.now()
			)

		self.cache.emoji.pop(cachekey, None)


	def status(self, handle, text, warning=None, domain=None):
		pass


	def user(self, handle, email, password, name=None, permissions=4):
		domain = self.fetch('domain', domain=config.domain)

		if not domain:
			logging.error(f'session.put.user: Unknown domain: {domain}')
			return

		row = self.fetch('user', handle=handle, domainid=domain.id)

		if row:
			logging.warning('db.functions.put.user: User already exists:', handle)
			return

		keys = GenKey()
		actor = f'https://{config.web_domain}/actor/{handle}'

		self.insert('user',
			handle = handle,
			lower = handle.lower(),
			name = name if name else handle,
			domainid = domainrow.id,
			email = email,
			password = mkhash(password),
			permissions = permissions,
			pubkey = keys.PUBKEY,
			privkey = keys.PRIVKEY,
			actor = actor,
			inbox = actor,
			config = user_config_defaults.copy(),
			timestamp = datetime.now()
		)


	def remote_user(self, actor: dict):
		print(actor)
		domain = self.fetch('domain', web_domain=actor.web_domain)
		handle = actor.preferredUsername

		if not domain:
			logging.error(f'session.put.user: Unknown domain: {web_domain}')
			return

		row = self.fetch('user', handle=handle, domainid=domain.id)

		if row:
			logging.warning('db.functions.put.user: User already exists:', handle)
			return

		self.insert('user',
			handle = handle,
			lower = handle.lower(),
			name = actor.get('name', handle),
			domainid = domain.id,
			permissions = 4 if actor.type == 'Person' else 0,
			pubkey = actor.pubkey,
			actor = actor.id,
			inbox = actor.inbox,
			config = user_config_defaults.copy(),
			timestamp = datetime.now()
		)



	def version(self, ver):
		self.config('version', ver, 'int')
