from IzzyLib.misc import DotDict
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, JSON, Boolean
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()
table = DotDict()

class Domain(Base):
	__tablename__ = 'domain'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	domain = Column('domain', String, nullable=False, unique=True)
	web_domain = Column('web_domain', String, nullable=False, unique=True)
	name = Column('name', String)
	admin = Column('admin', String)
	description = Column('description', String)
	short_description = Column('short_description', String)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'Domain(id={self.id}, domain="{self.domain}", name="{self.name}")'


class User(Base):
	__tablename__ = 'user'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	handle = Column('handle', String, nullable=False)
	lower = Column('lower', String, nullable=False)
	name = Column('name', String)
	domainid = Column('domainid', Integer, ForeignKey('domain.id'), nullable=False)
	email = Column('email', String)
	password = Column('password', String)
	bio = Column('bio', String)
	info_table = Column('info_table', JSON, nullable=False, default={})
	permissions = Column('permissions', Integer, default=4)
	privkey = Column('privkey', String)
	pubkey = Column('pubkey', String)
	enabled = Column('enabled', Boolean, default=True)
	silenced = Column('silenced', Boolean, default=False)
	suspended = Column('suspended', Boolean, default=False)
	actor = Column('actor', String, nullable=False)
	inbox = Column('inbox', String, nullable=False)
	config = Column('config', JSON, default={'private': False})
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'User(id={self.id}, handle="{self.handle}", name="{self.name}")'


class Status(Base):
	__tablename__ = 'status'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	userid = Column('userid', Integer, ForeignKey('user.id'))
	domainid = Column('domainid', Integer, ForeignKey('domain.id'))
	hash = Column('hash', String)
	content = Column('content', String)
	warning = Column('warning', String)
	visibility = Column('visibility', String, default='public')
	mentions = Column('mentions', String)
	replies = Column('replies', String)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'Status(id={self.id}, userid={self.userid}, hash="{self.hash}")'


class App(Base):
	__tablename__ = 'app'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	name = Column('name', String)
	url = Column('url', String)
	redirect_uri = Column('redirect_uri', String)
	client_id = Column('appid', String, nullable=False, unique=False)
	client_secret = Column('secret', String, nullable = False, unique=True)
	scope = Column('scope', JSON, nullable = False)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'App(id={self.id}, name="{self.name}", url="{self.url}")'


class Token(Base):
	__tablename__ = 'token'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	token = Column('token', String, unique=True)
	userid = Column('userid', Integer, ForeignKey('user.id'))
	appid = Column('appid', Integer, ForeignKey('app.id'))
	code = Column('code', String, unique=True)
	access = Column('access', DateTime)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'Code(id={self.id}, token="{self.token}")'


class Cookie(Base):
	__tablename__ = 'cookie'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	token = Column('token', String, nullable=False, unique=True)
	userid = Column('userid', Integer, ForeignKey('user.id'))
	address = Column('address', String)
	agent = Column('agent', String)
	access = Column('access', DateTime)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'Cookie(id={self.id}, cookie="{self.cookie}")'


class Emoji(Base):
	__tablename__ = 'emoji'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	shortcode = Column('shortcode', String, nullable=False)
	domainid = Column('domainid', Integer, nullable=False)
	display = Column('display', Boolean, default=True)
	enabled = Column('enabled', Boolean, default=True)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'Emoji(id={self.id}, shortcode="{self.shortcode}")'


class Object(Base):
	__tablename__ = 'object'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	uuid = Column('uuid', String)
	data = Column('data', JSON)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'Object(id={self.id}, uuid="{self.uuid}")'


class Config(Base):
	__tablename__ = 'config'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	key = Column('key', String, unique=True)
	value = Column('value', String)
	type = Column('type', String, default='str')


	def __repr__(self):
		return f'Config(id={self.id}, key="{self.key}", value="{self.value}", type="{self.type}")'


class Bans(Base):
	__tablename__ = 'bans'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	handle = Column('handle', String)
	domain = Column('domain', String)
	bantype = Column('bantype', String, default='silence')
	reject_media = Column('reject_media', Boolean, default=False)
	reason = Column('reason', String)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		if self.handle:
			return f'UserBan(id={self.id}, handle="{self.handle}", domain="{self.domain}", bantype="{self.bantype}", reason="{self.reason}")'
		else:
			return f'DomainBan(id={self.id}, domain="{self.domain}", bantype="{self.bantype}", reason="{self.reason}")'


class IPBlocks(Base):
	__tablename__ = 'ipblocks'

	id = Column('id', Integer, primary_key=True, autoincrement=True)
	address = Column('address', String, nullable=False)
	cidr = Column('cidr', Integer)
	timestamp = Column('timestamp', DateTime)


	def __repr__(self):
		return f'IPBlock(id={self.id}, address="{self.address}", cidr="{self.cidr}")'


table.update({
	'app': App,
	'config': Config,
	'cookie': Cookie,
	'emoji': Emoji,
	'domain': Domain,
	'object': Object,
	'status': Status,
	'token': Token,
	'user': User,
	'bans': Bans,
	'ipblocks': IPBlocks
})
