import functools, json, magic, os, re, sys, sanic, socket, yaml
import tarfile, tempfile, zipfile

from Crypto.Hash import SHA3_512, SHA3_256, SHA256, BLAKE2b
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from IzzyLib import logging, color
from IzzyLib.http import HttpClient, SetClient
from IzzyLib.misc import DefaultDict, DotDict
from base64 import b64decode, b64encode
from sanic import response
from datetime import datetime
from hashlib import md5
from markdown import markdown
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from .config import config

cache = DefaultDict()
css_check = lambda css_file : int(os.path.getmtime(config.frontend.join(f'{css_file}.css').str()))
cssts = DotDict()

hashalgs = {
	'blake': BLAKE2b,
	'sha256': SHA3_256,
	'sha512': SHA3_512,
	'md5': md5
}

paths = DotDict({
	'anon_api': [
		'/api/v1/apps',
		'/api/v1/instance'
	],
	'ap': [
		'/status',
		'/actor',
		'/inbox'
	],
	'cookie': [
		'/@',
		'/:',
		'/welcome',
		'/preferences',
		'/admin',
		'/oauth/authorize'
	]
})

themes = {
	'blue': {'primary': '#55e'},
	'green': {'primary': '#2e5'},
	'orange': {'primary': '#e52'},
	'red': {'primary': '#e24'},
	'pink': {'primary': '#e29'},
	'purple': {'primary': '#a2e'},
	'yellow': {'primary': '#ed2'}
}


def mkhash(string, salt=config.salt, alg='blake'):
	hashalg = hashalgs.get(alg.lower())

	if not hashalg:
		raise KeyError(f'Not a valid hash algorithm: {alg}')

	string = string.encode('UTF-8') if type(string) != bytes else string
	salt = salt.encode('UTF-8') if type(salt) != bytes else salt

	newhash = hashalg() if alg == 'md5' else hashalg.new()
	newhash.update(string)
	newhash.update(salt)
	return newhash.hexdigest()


def ApDate(dtobject):
	return dtobject.strftime('%Y-%m-%dT%H:%M:%SZ')


def CssTheme(theme):
	custpath = config.data.join(f'themes/{theme}.yml')
	defpath = config.frontend.join(f'themes/{theme}.yml')

	if custpath.isfile():
		themefile = custpath.str()

	elif defpath.isfile():
		themefile = defpath.str()

	else:
		themefile = None

	if cssfile:
		data = yaml.load(open(themefile, 'r'), Loader=yaml.FullLoader)

	else:
		data = {}

	colors = {
		'background': data.get('background', '#111'),
		'primary': data.get('primary', '#3bf'),
		'secondary': data.get('secondary', '#3bf'),
		
	}

	return colors


def CssTimestamp(*args):
	timestamp = 0

	for name in args:
		filename = config.frontend.join(f'{name}.css')
		timestamp += int(filename.mtime())

	return timestamp


#def timestamp(integer=True):
	#ts = datetime.timestamp(datetime.now())

	#return int(ts) if integer == True else ts


def Error(request, code, msg=None):
	message = msg if msg else ''

	if any(map(request.path.startswith, ['/status', '/actor'])) or JsonCheck(request):
		return JsonResp({'error': message}, code)

	else:
		cont_type = 'text/html'
		data = {
			'login_token': request.cookies.get('login_token') if not isinstance(request, str) else '',
			'code': str(code),
			'msg': msg
		}

		return config.template.response(f'error.haml', request, {'data': data, 'msg': message}, status=code)


def FetchActor(url, user=None):
	if user:
		actor = Client.signed_request(user.privkey, f'https://{config.web_domain}/actor/{user.handle}', url).json()

	else:
		actor = Client.request(url).json()

	actor.web_domain = urlparse(url).netloc
	actor.shared_inbox = actor.inbox
	actor.pubkey = None
	actor.handle = actor.preferredUsername

	if actor.get('endpoints'):
		actor.shared_inbox = actor.endpoints.get('sharedInbox', actor.inbox)

	if actor.get('publicKey'):
		actor.pubkey = actor.publicKey.get('publicKeyPem')

	return actor


def FetchApi(domain):
	return Client.request(f'https://{domain}/api/v1/instance').json()


@functools.lru_cache(maxsize=2048)
def FetchWebfingerAcct(handle, domain):
	data = DefaultDict({'actor': None})
	webfinger = Client.request(f'https://{domain}/.well-known/webfinger?resource=acct:{handle}@{domain}')

	if not webfinger.body:
		return

	data.handle, data.domain = webfinger.json().subject.replace('acct:', '').split('@')

	for link in webfinger.json().links:
		if link['rel'] == 'self' and link['type'] == 'application/activity+json':
			data.actor = link['href']

	return data


def ForwardCheck(header):
	forwarded = DotDict(valid=False)

	for pair in header.split(';'):
		k,v = pair.split('=', 1)

		if k == 'for':
			k = 'ip'

		forwarded[k] = v

	if forwarded.secret == config.secret:
		forwarded.valid = True

	return forwarded


def GenKey():
	privkey = RSA.generate(2048)
	pubkey = privkey.publickey()

	return DotDict({
		'pubkey': pubkey,
		'privkey': privkey,
		'PUBKEY': pubkey.exportKey('PEM').decode(),
		'PRIVKEY': privkey.exportKey('PEM').decode()
	})


def GetIp():
	s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

	try:
		s.connect(('10.255.255.255', 1))
		data = s.getsockname()
		ip = data[0]

	except Exception:
		ip = '127.0.0.1'

	finally:
		s.close()

	return ip


def GetTheme(name):
	defaults = {
		'primary': '#e7a',
		'secondary': '#e7a',
		'background': '#191919',
		'positive': '#ada',
		'negative': '#daa'
	}

	defaults.update(themes.get(name, {}))
	return defaults


def JsonCheck(request):
	if any(map(request.path.startswith, ['/actor', '/inbox', '/status'])) or request.path.endswith('.json'):
		return True

	accept = request.headers.get('Accept')

	if not accept:
		return

	mimes = accept.split(',')

	if any(mime in ['application/json', 'application/activity+json'] for mime in mimes):
		return True

	return False


def JsonResp(data, status=200, headers=None, activity=False):
	params = {
		'content_type': 'application/activity+json' if activity else 'application/json',
		'status': status
	}

	if headers:
		params['headers'] = headers

	return sanic.response.text(json.dumps(data), **params)


def ParseRequest(request, db):
	request.ctx.parsed = True
	request.ctx.media = False
	request.ctx.query = DotDict(dict(request.query_args))
	request.ctx.form = DotDict({k:v[0] for k,v in request.form.items()})
	request.ctx.files = DotDict({k:v[0] for k,v in request.files.items()})

	if any(map(request.path.startswith, ['/static', '/media', '/favicon'])):
		request.ctx.media = True
		return

	request.ctx.api_path = request.path.startswith('/api')
	request.ctx.api_anon_path = any(map(request.path.startswith, paths.anon_api))
	request.ctx.ap_path = any(map(request.path.startswith, paths.ap))
	request.ctx.cookie_path = any(map(request.path.startswith, paths.cookie))
	request.ctx.json_path = JsonCheck(request)

	with db.session() as s:
		request.ctx.token = None
		request.ctx.token_user = None
		request.ctx.cookie = None
		request.ctx.cookie_user = None

		api_token = request.headers.get('authorization', '').replace('Bearer ', '')
		login_token = request.cookies.get('login_token')

		if api_token:
			api_db = s.get.token(api_token, 'token')
			request.ctx.token = api_db['token']
			request.ctx.token_user = api_db['user']

		if login_token:
			login_db = s.get.token(login_token, 'cookie')
			request.ctx.cookie = login_db['token']
			request.ctx.cookie_user = login_db['user']

def ParseSig(signature):
	if not signature:
		logging.verbose('Missing signature header')
		return

	split_sig = signature.split(',')
	sig = DefaultDict({})

	for part in split_sig:
		key, value = part.split('=', 1)
		sig[key.lower()] = value.replace('"', '')

	if not sig.headers:
		logging.verbose('Missing headers section in signature')
		return

	sig.headers = sig.headers.split()

	return sig


def ParseText(text, style='markdown'):
	'insert code to process text here'
	if style == 'markdown':
		return markdown(text)


def PkcsHeaders(key, headers, sig=None):
	if sig:
		head_items = [f'{item}: {headers[item]}' for item in sig.headers]

	else:
		head_items = [f'{k.lower()}: {v}' for k,v in headers.items()]

	head_string = '\n'.join(head_items)
	head_bytes = head_string.encode('UTF-8')

	KEY = RSA.importKey(key)
	pkcs = PKCS1_v1_5.new(KEY)
	h = SHA256.new(head_bytes)

	if sig:
		return pkcs.verify(h, b64decode(sig.signature))

	else:
		return pkcs.sign(h)


def StrDate(ts):
	return datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M')


def Themes():
	theme_list = []

	for theme in os.listdir(f'{config.path}/themes'):
		theme_list.append(theme.replace('.yml', ''))

	theme_list.sort()

	return theme_list


#def VerifyString(string, enc_string, pubkey, alg='SHA256'):
	#KEY = RSA.importKey(key)
	#pkcs = PKCS1_v1_5.new(KEY)
	#h = SHA256.new(string.encode('UTF-8'))
	#return pkcs.verify(h, b64decode(enc_string))


#class HttpClient(object):
	#def __init__(self, headers={}, useragent='IzzyLib/0.3', proxy_type='https', proxy_host=None, proxy_port=None):
		#proxy_ports = {
			#'http': 80,
			#'https': 443
		#}

		#if proxy_type not in ['http', 'https']:
			#raise ValueError(f'Not a valid proxy type: {proxy_type}')

		#self.headers=headers
		#self.agent=useragent
		#self.proxy = DotDict({
			#'enabled': True if proxy_host else False,
			#'ptype': proxy_type,
			#'host': proxy_host,
			#'port': proxy_ports[proxy_type] if not proxy_port else proxy_port
		#})


	#def __sign_request(self, request, privkey, keyid):
		#request.add_header('(request-target)', f'{request.method.lower()} {request.path}')
		#request.add_header('host', request.host)
		#request.add_header('date', datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT'))

		#if request.body:
			#body_hash = b64encode(SHA256.new(request.body).digest()).decode("UTF-8")
			#request.add_header('digest', f'SHA-256={body_hash}')
			#request.add_header('content-length', len(request.body))

		##head_items = [f'{k.lower()}: {v}' for k,v in request.headers.items()]
		##head_string = '\n'.join(head_items)

		##PRIVKEY = RSA.importKey(privkey)
		##pkcs = PKCS1_v1_5.new(PRIVKEY)
		##h = SHA256.new(head_string.encode('UTF-8'))

		#sig = {
			#'keyId': keyid,
			#'algorithm': 'rsa-sha256',
			#'headers': ' '.join([k.lower() for k in request.headers.keys()]),
			#'signature': b64encode(PkcsHeaders(privkey, request.headers)).decode('UTF-8')
		#}

		#sig_items = [f'{k}="{v}"' for k,v in sig.items()]
		#sig_string = ','.join(sig_items)

		#request.add_header('signature', sig_string)

		#request.remove_header('(request-target)')
		#request.remove_header('host')


	#def __build_request(self, url, data=None, headers={}, method='GET'):
		#new_headers = self.headers.copy()
		#new_headers.update(headers)

		#parsed_headers = {k.lower(): v.lower() for k,v in new_headers.items()}

		#if not parsed_headers.get('user-agent'):
			#parsed_headers['user-agent'] = self.agent

		#if isinstance(data, dict):
			#data = json.dumps(data)

		#if isinstance(data, str):
			#data = data.encode('UTF-8')

		#request = HttpRequest(url, data=data, headers=parsed_headers, method=method)

		#if self.proxy.enabled:
			#request.set_proxy(f'{self.proxy.host}:{self.proxy.host}', self.proxy.ptype)

		#return request


	#def request(self, *args, **kwargs):
		#request = self.__build_request(*args, **kwargs)

		#try:
			#response = urlopen(request)
		#except HTTPError as e:
			#response = e.fp

		#return HttpResponse(response)


	#def json(self, *args, headers={}, activity=True, **kwargs):
		#json_type = 'activity+json' if activity else 'json'
		#headers.update({
			#'accept': f'application/{json_type}'
		#})
		#return self.request(*args, headers=headers, **kwargs)


	#def signed_request(self, privkey, keyid, *args, **kwargs):
		#request = self.__build_request(*args, **kwargs)

		#self.__sign_request(request, privkey, keyid)

		#try:
			#response = urlopen(request)
		#except HTTPError as e:
			#response = e

		#return HttpResponse(response)


#class HttpRequest(Request):
	#def __init__(self, *args, **kwargs):
		#super().__init__(*args, **kwargs)
		#parsed = urlparse(self.full_url)

		#self.scheme = parsed.scheme
		#self.host = parsed.netloc
		#self.domain = parsed.hostname
		#self.port = parsed.port
		#self.path = parsed.path
		#self.query = parsed.query
		#self.body = self.data if self.data else b''


#class HttpResponse(object):
	#def __init__(self, response):
		#self.body = response.read()
		#self.headers = DefaultDict({k.lower(): v.lower() for k,v in response.headers.items()})
		#self.status = response.status
		#self.url = response.url


	#def text(self):
		#return self.body.decode('UTF-8')


	#def json(self):
		#return DotDict(self.text())


	#def json_pretty(self, indent=4):
		#return json.dumps(self.json().asDict(), indent=indent)


proxy_config = {}

if config.proxy.enabled:
	proxy_config = config.proxy.asDict()
	del proxy_config['enabled']

Client = HttpClient(useragent=config.agent, **proxy_config)
SetClient(Client)
