import binascii, base64, json, hashlib, multiprocessing, traceback

from urllib.parse import quote, urlparse, unquote

from IzzyLib import logging
from IzzyLib.cache import LRUCache
from IzzyLib.http import ParseSig, PkcsHeaders, VerifyRequest
from IzzyLib.misc import DefaultDict, DotDict
from Crypto.Hash import SHA, SHA256, SHA512
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from base64 import b64encode, b64decode
from sanic.response import raw, redirect, json as rjson
from sanic import exceptions as exc
from tldextract import extract

from .config import config
from .database import db
from .errors import Teapot
from .functions import Client, FetchActor, FetchWebfingerAcct, ForwardCheck, ParseRequest


blocked_agents = [
	'gabsocial',
	'soapbox',
	'spinster',
	'kiwifarms'
]

domain_bans = [
	'b77625108b127dc1b15640ac536c8e2991b83ea20841a42e70927f96434fe9c2',
	'f1e9527ea188da343fe79f23855f80476cedb017e3880432ab22467eaf992f44',
	'a67a3f85d90881ae32716d0eaf5a348b53096e07ea38a20797d74824074e281b',
	'75e80dd194bc635186cf0b4a07524d7746b4b261cbd5fb15007189e4fccfbbb6',
	'56638777979b0e88cd6cd67723f4fc152eae7ada5c266cdab45e8bb29af6be24',
	'acb2a6826ab0741142935b0e0454b381c0b32a5dfcf54bd43c354379263aec1f',
	'e249dc1503a399ad9fb807ded9be6dd42ca9ea04fa04adb8f794435e1ddb8554',
	'feea7dc93ff4b6b7c06bc3d2dbd6868b259dccd664309c859ed56d3480971617',
	'f0f7b4b28f4142e276d39b1fddfe621e81f72d3be86296a6323d303164f423a2',
	'bfaa1ed5a206867a79f7fbd187f98e035a1cee5fa1e3821740dfc0a9cb3d6ecc',
	'0dbb3144d7b6dff976d97a54be2e94da3559d089eb5c00f871e18318afb4f047',
	'56da8b2c7030ff79b9acba7a23fb6750a165a88505e0dcac85478d537d88e90f',
	'6dd58fd367069c428c481182a8a6f62af21b7a33207b12cb15140da06b8401af',
	'e7c641cf9cd8b5c4513e22d0dfd3f914835a16bd4011801dfa0976fedf61bf02',
	'a3da30d2dfb76009289509cb3f35abd0a43b0736f0abf1ddd17c05685902a6fb',
	'19452ff02670099643197b3fec15c0d43e42a3dbbe83b0fb4d039d5f77d1932e',
	'e8bd7afb2125b7ea5e1b885fac2a657d817eec289ed870aa3d1d0acbe6aded8f'
]

blocked_domains = [extract(domain) for domain in domain_bans]
hash_cache = LRUCache(4096)

log_ignore_paths = [
	'/media',
	'/static',
	'/style',
	'/manifest',
	'/favicon'
]


def CSP():
	host = config['web_domain']
	data = ''
	params = {
		'default-src': ['self', host],
		'script-src': ['self', 'unsafe-inline', host],
		'connect-src': ['self', 'blob:', 'data:', f'wss://{host}'],
		'frame-src': ['self', 'https:'],
		'frame-ancestors': ['none'],
		'object-src': ['none']
	}

	#params = {
		#'base-uri': 'none',
		#'default-src': f'self {host}',
		#'frame-ancestors': 'none',
		##'script-src': f'self {host}',
		##'font-src': f'self {host}',
		#'img-src': f'self data: blob: {host}',
		#'style-src': f'self unsafe-inline {host}',
		#'media-src': f'self data: {host}',
		#'frame-src': 'self https:',
		#'worker-src': f'self blob: {host}',
		#'connect-src': f'self blob: data: wss://{host} {host}',
		##'manifest-src': f'self {host}'
	#}

	for k,v in params.items():
		items = []

		for item in v:
			if item in ['self', 'unsafe-inline', 'none']:
				items.append(f"'{item}'")
			else:
				items.append(item)

		value = ' '.join(items)
		data += f"{k} {value}; "

	return data


csp_header = CSP()


async def http_domain_user_check(request):
	ParseRequest(request, db)

	request.ctx.actor = None
	request.ctx.domain = None
	request.ctx.web_domain = None
	request.ctx.top_domain = None
	request.ctx.valid = None

	if request.ctx.media:
		return

	if not request:
		logging.warning('Empty request')
		raise exc.InvalidUsage('Empty request')

	request.ctx.signature = ParseSig(request.headers.get('signature'))

	if not request.ctx.signature:
		return

	keyid = request.ctx.signature.keyid
	web_domain = urlparse(keyid).netloc
	domain_extract = extract(keyid)
	top_domain = f'{domain_extract.domain}.{domain_extract.suffix}'
	domain_cache = hash_cache.get(top_domain)

	request.ctx.actor_url = keyid.split('#', 1)[0]
	request.ctx.valid = VerifyRequest(request)
	request.ctx.top_domain = top_domain
	request.ctx.web_domain = web_domain

	actor = FetchActor(request.ctx.actor_url, db.fetch('user', handle='system'))
	webfinger = FetchWebfingerAcct(actor.preferredUsername, web_domain)

	if not webfinger:
		return

	request.ctx.domain = webfinger.domain
	request.ctx.actor = actor

	with db.session() as s:
		domainrow = s.fetch('domain', domain=webfinger.domain)

		if not domainrow:
			s.put.domain(webfinger.domain, web_domain)
			domainrow = s.fetch('domain', domain=webfinger.domain)
			s.commit()

		if not s.get.user(actor.preferredUsername, webfinger.domain):
			s.put.remote_user(actor)
			actor = s.get.user(webfinger.handle, webfinger.domain)

		request.ctx.domainrow = domainrow
		request.ctx.actorrow = actor


async def http_bans(request):
	# Let the request go through if it's basic resources needed for the error page
	if any(map(request.path.startswith, ['/favicon', '/style', '/static'])):
		return

	if [agent for agent in blocked_agents if agent in request.headers.get('User-Agent', '').lower()]:
		raise Teapot('This teapot kills fascists')

	if not request.ctx.top_domain:
		return

	domain_hash = hash_cache.fetch('domain')

	if not domain_hash:
		domain = request.ctx.top_domain
		domain_hash = hashlib.sha256(domain.encode("utf-8")).hexdigest()
		hash_cache.store(domain, domain_hash)

	if domain_hash in blocked_domains:
		raise Teapot('This teapot kills fascists')


async def http_auth(request):
	if request.ctx.media:
		return

	if not request.ctx.api_anon_path and request.ctx.api_path:
		resp = handle_api_path(request)

		if resp:
			return resp

	elif request.ctx.ap_path:
		resp = handle_api_path(request)

		if resp:
			return resp

	elif request.ctx.cookie_path:
		resp = handle_api_path(request)

		if resp:
			return resp


async def http_headers(request, response):
	#always_cache = ['ico', 'css', 'svg', 'js', 'png', 'woff2']

	prod_cache_ext = ['css', 'svg', 'js', 'ico', 'jpg', 'png', 'woff2', 'mp4', 'mov', 'webm']
	prod_cache_path = ['/manifest.json']
	is_prod = config.env == 'prod'

	raw_ext = request.path.split('.')[-1:]
	ext = raw_ext[0] if len(raw_ext) > 0 else None

	if request.ctx.parsed and not request.ctx.media:
		if request.ctx.ap_path:
			response.headers['Content-Type'] = 'application/activity+json'

		elif request.ctx.api_path or request.ctx.json_path:
			response.headers['Content-Type'] = 'application/json'

	if not response.headers.get('Cache-Control'):
		compare = [
			ext in prod_cache_ext,
			request.path.startswith('/media'),
			request.path in prod_cache_path
		]

		if is_prod and any(compare):
			response.headers['Cache-Control'] = 'public,max-age=2628000,immutable'

		else:
			response.headers['Cache-Control'] = 'no-store'

	response.headers['Content-Security-Policy'] = csp_header
	response.headers['Server'] = f'BarksharkSocial/{config.version}'
	response.headers['Trans'] = 'Rights'


async def http_access_log(request, response):
	if config.env == 'dev' and (any(map(request.path.startswith, log_ignore_paths)) or response.status == 302):
		return

	uagent = request.headers.get('user-agent')
	forwarded = ForwardCheck(request.headers.get('forwarded'))
	address = request.remote_addr if not forwarded.valid else forwarded.ip

	logging.info(f'({multiprocessing.current_process().name}) {address} {request.method} {request.path} {response.status} "{uagent}"')


def handle_web_path(request):
	with db.session() as s:
		handle = request.match_info.get('user')
		statusid = request.match_info.get('statusid')

		if any(map(request.path.startswith, ['/@', '/:'])) and request.ctx.json_path:
			return

		if request.path.startswith('/@'):
			domainid = s.get.domainid(config.domain)
			user = s.fetch('user', handle=handle, domainid=domainid)

			if user.config.get('public'):
				return

	if not request.ctx.cookie:
		if request.ctx.json_path:
			raise exc.Unauthorized('No Token')

		else:
			return redirect(f'/login?redir={quote(request.path)}&query={quote(request.query_string)}')

	elif not request.ctx.cookie:
		return redirect(f'/login?redir={quote(request.path)}&msg=InvalidToken')


def handle_api_path(request):
	if not request.ctx.token or not request.ctx.token_user:
		logging.debug('middleware.http_auth: Invalid token:', request.ctx.token)
		return response.json({'error': 'Invalid token'}, status=401)


def handle_ap_path(request):
	if not request.ctx.valid:
		logging.debug('middleware.http_auth: Signature verify failed')

		if request.method == 'GET' and request.path.startswith('/actor') and len(request.path.split('/')) == 3:
			pass

		else:
			raise exc.Unauthorized('Failed to verify signature')

	else:
		request.ctx.valid = True


#def ValidateRequest(request):
	#headers = {k.lower(): v for k,v in request.headers.items()}
	#headers['(request-target)'] = f'{request.method.lower()} {request.path}'
	#signature = request.ctx.signature
	#actor = request.ctx.actor
	#digest = headers.get('digest')

	#if not signature:
		#return False

	### Make sure there is a digest header for inbox pushes
	#if request.method.lower() == 'post' and (request.path.startswith('/actor') or request.path == '/inbox') and not request.body:
		#return False

	### Date and Host headers are required
	#elif None in (headers.get('date'), headers.get('host')):
		#return False

	### 
	#elif request.body and digest and not ValidateBody(digest, request.body):
		#return False

	#try:
		#resp = Client.json(actor)
		#actor_data = resp.json()
		#pubkey = actor_data.publicKey['publicKeyPem']

	#except Exception as e:
		#traceback.print_exc()
		#logging.verbose(f'Failed to get public key for actor {signature.keyid}')
		#return False

	#request.ctx.valid = PkcsHeaders(pubkey, {k:v for k,v in headers.items() if k in signature.headers}, sig=signature)

	#return request.ctx.valid


class InvalidAttr(object):
	pass
