import json

from sanic.views import HTTPMethodView
from sanic import response

from .. import messages
from ..config import config
from ..functions import Error, JsonResp
from ..database import db


class Status(HTTPMethodView):
	async def get(self, request, statusid, activity=None):
		data = response.json(messages.Note(statusid))

		if activity:
			return messages.Create(data)

		return data


class Replies(HTTPMethodView):
	async def get(self, request, statusid=None):
		data = {'msg': 'UvU'}
		return response.json(data)


class Actor(HTTPMethodView):
	async def get(self, request, user=None, extra=None):
		user = db.get().user(user)
		resp = {
			'following': self._following,
			'followers': self._followers,
			'collections': self._collections,
			'featured': self._featured,
			'outbox': self._outbox
		}

		#print(user)

		command = resp.get(extra)

		if user == None:
			return error(request, 404, 'That user doesn\'t exist.')

		if not command:
			if extra:
				return error(request, 404, 'This user data doesn\'t exist.')

			else:
				data = self._actor_data(request, user)

		else:
			data = command(request, user)

		return response.json(data)


	async def post(self, request, user=None, extra=None):
		with db.session() as s:
			user = s.fetch('user', handle=user)

		data = {}
		print(json.dumps(request.json, indent=4))
		return response.json(data, status=500)


	def _actor_data(self, request, user):
		return messages.Actor(request, user)


	def _following(self, request, user):
		data = {'msg': 'UvU'}
		return data


	def _followers(self, request, user):
		data = {'msg': 'UvU'}
		return data


	def _collections(self, request, user):
		data = {'msg': 'UvU'}
		return data


	def _featured(self, request, user):
		data = {'msg': 'UvU'}
		return data


	def _outbox(self, request, user):
		outbox = f'https://{config.web_domain}/actor/{user.handle}/outbox'

		with db.session() as s:
			status = s.get.statuses('izalia')
			Status = s.classes.Status

			if request.ctx.query.get('page'):
				try:
					page_num = int(request.ctx.query.get('page'))
				except TypeError:
					return {'error': 'not a valid number'}

				page = status.page(page_num)
				data = {
					'@context': [
						'https://www.w3.org/ns/activitystreams',
						{
							'ostatus': 'http://ostatus.org#',
							'atomUri': 'ostatus:atomUri',
							'inReplyToAtomUri': 'ostatus:inReplyToAtomUri',
							'conversation': 'ostatus:conversation',
							'sensitive': 'as:sensitive',
							'toot': 'http://joinmastodon.org/ns#',
							'votersCount': 'toot:votersCount',
							'blurhash': 'toot:blurhash',
							'focalPoint': {
								'@container': '@list',
								'@id': 'toot:focalPoint'
							}
						}
					],
					'id': f'{outbox}?page=true',
					'type': 'OrderedCollectionPage',
					'partOf': f'{outbox}'
				}

				if page.has_next():
					data['next'] = f'{outbox}?page={page.next_page_number}'

				if page.has_previous():
					data['prev'] = f'{outbox}?page={page.previous_page_number}'

				data['orderedItems'] = [messages.Note(row.id, Status(row, db)) for row in page.object_list]

			else:
				count = s.count('status', userid=user.id)
				data = {
					'@context': 'https://www.w3.org/ns/activitystreams',
					'id': outbox,
					'type': 'OrderedCollection',
					'totalItems': count,
					'first': f'{outbox}?page=1',
					'last': f'{outbox}?page='
				}

			return data


class Inbox(HTTPMethodView):
	async def post(self, request):
		print(request.body)
		return response.json({'msg': 'UvU'}, content_type='application/activity+json', status=202)


class Nodeinfo(HTTPMethodView):
	async def get(self, request):
		with db.session() as s:
			settings = s.get.configs()
			stats = s.get.stats()

		data = {
			'version': '2.0',
			'usage': {
				'users': {
					'total': stats.user_count,
					'activeMonth': stats.users_month,
					'activeHalfyear': stats.users_halfyear
				},
				'localPosts': stats.status_count
			},
			'software': {
				'name': 'bsocial',
				'version': config.version
			},
			'protocols': ['activitypub'],
			'openRegistrations': False,
			'metadata': {
				'staffAccounts': [],
				'postFormats': ['text/plain', 'text/html', 'text/markdown'],
				'nodeName': settings.name,
				'nodeDescription': settings.description
			}
		}

		return response.json(data)


class Wellknown(HTTPMethodView):
	async def get(self, request, name):
		endpoints = {
			'nodeinfo': self.NodeInfo,
			'host-meta': self.HostMeta,
			'webfinger': self.WebFinger
		}

		endpoint = endpoints.get(name)

		if not endpoint:
			return error(request, 404, 'Invalid well-known endpoint')

		ctype, data = endpoint(request)

		if ctype.endswith('json'):
			return response.json(data)

		if name == 'webfinger' and not data:
			return JsonResp({}, 404)

		return response.text(data, content_type=ctype)


	def NodeInfo(self, request):
		data = {
			'links': [
				{
					'rel': 'http://nodeinfo.diaspora.software/ns/schema/2.0',
					'href': f'https://{config.web_domain}/nodeinfo/2.0.json'
				}
			]
		}

		return ('application/json', data)


	def HostMeta(self, request):
		data = f'''<?xml version="1.0" encoding="UTF-8"?>
<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">
	<Link rel="lrdd" type="application/xrd+xml" template="https://{config.web_domain}/.well-known/webfinger?resource={{uri}}"/>
</XRD>
'''

		return ('application/xrd+xml', data)


	def WebFinger(self, request):
		resource_query = request.ctx.query.get('resource')

		if not resource_query:
			return ('application/json', {})

		resource = resource_query.split('@')
		handle = resource[0].replace('acct:', '')
		user = db.fetch('user', handle=handle)

		if resource[1] == config.domain and user:
			data = {
				'subject': f'acct:{handle}@{config.domain}',
				'aliases': [
					f'https://{config.web_domain}/@{handle}',
					f'https://{config.web_domain}/actor/{handle}'
				],
				'links': [
					{
						'rel': 'http://webfinger.net/rel/profile-page',
						'type': 'text/html',
						'href': f'https://{config.web_domain}/@{handle}'
					},
					{
						'rel': 'self',
						'type': 'application/activity+json',
						'href': f'https://{config.web_domain}/actor/{handle}'
					},
					{
						'rel': 'http://ostatus.org/schema/1.0/subscribe',
						'template': f'https://{config.web_domain}/interact?url={{url}}'
					}
				]
			}

			return ('application/json', data)

