import os, validators

from IzzyLib.http import VerifyString
from datetime import datetime
from sanic import response
from sanic.views import HTTPMethodView
from urllib.parse import unquote

from .. import settings, messages
from ..config import config
from ..database import db
from ..functions import Error, GenKey, mkhash


class Home(HTTPMethodView):
	async def get(self, request):
		return config.template.response('pages/home.haml', request, {})


class Rules(HTTPMethodView):
	async def get(self, request):
		return config.template.response(f'pages/rules.haml', request, {})


class About(HTTPMethodView):
	async def get(self, request):
		with db.session() as s:
			data = {
				'about': s.get.config('about'),
				'admins': s.fetch('user', permissions=1, single=False),
				'mods': s.fetch('user', permissions=2, single=False)
			}
		return config.template.response('pages/about.haml', request, data)


class User(HTTPMethodView):
	async def get(self, request, user=None):
		if request.ctx.json_path:
			return response.json(messages.Actor(request, user))

		if not user:
			return Error(request, 404, 'User not specified')

		if '@' in user:
			handle, domain = user.split('@', 1)

		else:
			handle = user
			domain = config.domain

		if domain == config.domain and handle in ['system', 'relay', 'local']:
			return Error(request, 404, 'User not found')

		with db.session() as s:
			user = s.get.user(user, domain)
			page_num = request.ctx.query.get('page', 1)

			if not user:
				return Error(request, 404, 'User not found')

			page_data = s.get.statuses(user=user)

			if not page_data:
				posts = []

			else:
				page = page_data.page(page_num)
				posts = [db.classes.Status(row, db) for row in page.object_list]

		info_table = {}

		for k,v in user.info_table.items():
			if validators.url(v):
				v = f"<a href='{v}'>{v}</a>"

			info_table[k] = v

		user.info_table = info_table
		user.bio = user.bio.replace('\n', '<br>\n') if user.bio else None

		context = {'user': user, 'posts': posts, 'page_data': page}
		return config.template.response('pages/profile.haml', request, context)


class Status(HTTPMethodView):
	def get(self, request, statusid=None):
		if request.ctx.json_path:
			if not request.ctx.valid:
				return response.json({'error': 'Request not signed or signature not valid'}, status=401)

			return response.json(messages.Note(statusid))

		with db.session() as s:
			status = s.fetch('status', id=statusid)

		if not status:
			return Error(request, 404, 'Status not found')

		return config.template.response('pages/status.haml', request, {'status': status})


class Register(HTTPMethodView):
	async def get(self, request, msg=None):
		return config.template.response('pages/register.haml', request, {'msg': msg})


	async def post(self, request):
		data = request.ctx.form

		pass1 = data.get('newpassword1')
		pass2 = data.get('newpassword2')

		username = data.get('username')
		password = pass1 if None not in [pass1, pass2] and pass1 == pass2 else None
		email = data.get('email')
		name = data.get('name')
		#sig = data.get('sig')
		#info_table = data.get('table')
		address = request.headers.get('X-Real-Ip')
		agent = request.headers.get('User-Agent')

		with db.session() as s:
			user = s.fetch('user', handle=username)

			if user:
				return await self.get(request, 'User already exists')

			if None in [username, password]:
				return await self.get(request, 'Missing username or password. Try again')

			s.put.user(username,
				name = name,
				email = email,
				password = mkhash(password),
			)

			user = s.fetch('user', handle=username)
			token = s.put.cookie(user.id, address, agent)

		resp = response.redirect('/')
		resp['login_token'] = token
		resp['login_token']['max-age'] = 60*60*24*14

		return resp


# login page (/login)
# I wanna add keypair login, but apparently string signing via client-side js is a fuck
class Login(HTTPMethodView):
	async def get(self, request, msg=None):
		query = request.ctx.query
		page = 'login_keypair' if query.get('login_type') == 'keypair' else 'login'
		response = config.template.response(f'pages/{page}.haml', request, {'msg': msg, 'redir': {'path': query.get('redir'), 'data': query.get('query')}})

		if not request.ctx.cookie_user and request.cookies.get('login_token'):
			del response.cookies['login_token']

		return response


	async def post(self, request):
		query = request.ctx.query
		headers = request.headers
		data = request.ctx.form

		login_type = data.get('login_type')
		username = data.get('username')

		with db.session() as s:
			if username:
				user = s.get.user(username)

				if not user:
					return await self.get(request, 'User not found')

			if data.get('login_type') == 'password':
				password = data.get('password')
				address = headers.get('X-Real-Ip')
				agent = headers.get('User-Agent')
				redir = data.get('redir')
				redir_data = data.get('redir_data')

				if None in [username, password]:
					return await self.get(request, 'Username or password is missing')

				if username in ['system', 'relay']:
					return await self.get(request, 'Cannot login to a system account')

				if user.password != mkhash(password):
					return await self.get(request, 'Password is invalid')

			elif login_type == 'keypair':
				login_value = data.get('login_value')
				encrypt_value = data.get('encrypt_value')
				pubkey = user.config.get('pubkey')

				if not pubkey:
					return await self.get(request, 'No saved pubkey')

				if not VerifyString(login_value, encrypt_value, pubkey):
					return await self.get(request, 'Failed to verify login value')

			else:
				return await self.get(request, f'Invalid login type: {login_type}')

			login_token = s.put.cookie(user.id, request.headers.get('X-Real-Ip'), request.headers.get('user-agent'))
			resp = response.redirect(f'{redir}?{unquote(redir_data)}' if redir and redir != 'None' else '/')

			request.ctx.cookie = login_token
			request.ctx.cookie_user = user

			# Send login token. Lasts for 2 weeks (60 sec * 60 min * 24 hour * 14 day)
			resp.cookies['login_token'] = login_token.token
			resp.cookies['login_token']['max-age'] = 60*60*24*14

			return resp


class Logout(HTTPMethodView):
	async def get(self, request):
		token = request.ctx.cookie

		if token:
			token.delete()

		request.ctx.cookie = None
		request.ctx.cookie_user = None

		resp = await Login().get(request, msg='Logged out' if token else None)
		del resp.cookies['login_token']
		return resp


class Preferences(HTTPMethodView):
	async def get(self, request, name=None, message=None, error=None):
		if not name:
			return response.redirect('/settings/profile')

		func = getattr(settings, f'get_user_{name}')
		data = func(request)
		data['message'] = message
		data['error'] = error

		if type(data) != dict:
			return data

		return config.template.response(f'pages/preferences/{name}.haml', request, data)


	async def post(self, request, name=None):
		if not name:
			return Error(request, 404, 'Not found')

		func = getattr(settings, f'post_user_{name}')
		data = func(request)
		message = None
		error = None

		if type(data) == dict:
			error = data.get('error')
			message = data.get('message')

		if data and type(data) not in [str, dict]:
			return data

		return await self.get(request, name=name, message=message, error=error)


class Admin(HTTPMethodView):
	async def get(self, request, name=None, message=None, error=None):
		if not name:
			return response.redirect('/admin/dashboard')

		func = getattr(settings, f'get_admin_{name}')
		data = func(request)
		data['message'] = message
		data['error'] = error

		if type(data) != dict:
			return data

		return config.template.response(f'pages/admin/{name}.haml', request, data)


	async def post(self, request, name=None):
		if not name:
			return Error(request, 404, 'Not found')

		func = getattr(settings, f'post_admin_{name}')
		data = func(request)
		message = None
		error = None

		if type(data) == dict:
			error = data.get('error')
			message = data.get('message')

		if data and type(data) not in [str, dict]:
			return data

		return await self.get(request, name=name, message=message, error=error)
