import asyncio, signal, sys, json

import sanic

from IzzyLib import logging
from IzzyLib.misc import GetIp
from sanic import exceptions as exc
from sanic_cors import CORS
from jinja2.exceptions import TemplateNotFound
from sqlalchemy.orm.exc import DetachedInstanceError
from uuid import uuid4

from . import views, errors, middleware
from .database import db
from .config import config
from .functions import CssTimestamp, StrDate, Themes
from .errors import Teapot


config.template.updateEnv({
	'config': config,
	'db': db,
	'uuid': uuid4,
	'CssTimestamp': CssTimestamp,
	'StrDate': StrDate,
	'Themes': Themes,
	'json': json
})

web = sanic.Sanic('social')
web.config['FORWARDED_SECRET'] = config.secret
web.config['API_BASEPATH'] = '/apidoc'
cors = CORS(web, resources={
	r'/*': {'origins': config['web_domain']},
	r'/media/*': {'origins': '*'},
	r'/api/*': {'origins': '*'},
	f'/oauth/*': {'oritins': '*'}
})


def safe_stop(*args):
	#logging.debug('Saving updated timestamps')
	#update_timestamps()

	logging.info('Bye')
	sys.exit()


def setupRoutes():
	error_handlers = [
		(exc.NotFound, errors.generic),
		(exc.MethodNotSupported, errors.generic),
		(exc.InvalidUsage, errors.generic),
		(exc.Unauthorized, errors.generic),
		(exc.Forbidden, errors.generic),
		(Teapot, errors.generic),
		(TemplateNotFound, errors.missing_template),
		(DetachedInstanceError, errors.orm_detached_instance),
		(Exception, errors.server_error)
	]

	for handler in error_handlers:
		web.error_handler.add(*handler)

	middlewares = [
		(middleware.http_domain_user_check, 'request'),
		(middleware.http_bans, 'request'),
		(middleware.http_headers, 'response'),
		(middleware.http_access_log, 'response')
	]

	for m, a in middlewares:
		web.register_middleware(m, attach_to=a)

	web_routes = [
		(views.api.Oauth, '/oauth/<name>'),
		(views.api.MastodonBase, '/api/v1/<path:[A-z0-9\/]+>'),
		(views.activitypub.Actor, '/actor/<user>'),
		(views.activitypub.Actor, '/actor/<user>/<extra>'),
		(views.activitypub.Status, '/status/<statusid>'),
		#(views.activitypub.Actor, '/actor/<user>'),
		#(views.activitypub.Inbox, '/inbox'),
		(views.activitypub.Nodeinfo, '/nodeinfo/2.0.json'),
		(views.activitypub.Wellknown, '/.well-known/<name>'),
		#(views.activitypub.Status, '/status/<status>'),
		(views.frontend.Home, '/'),
		(views.frontend.About, '/about'),
		(views.frontend.Rules, '/about/rules'),
		(views.frontend.Register, '/register'),
		(views.frontend.Login, '/login'),
		(views.frontend.Logout, '/logout'),
		(views.frontend.User, '/@<user>'),
		(views.frontend.Status, '/:<statusid:int>'),
		#(views.frontend.Welcome, '/welcome'),
		(views.frontend.Preferences, '/preferences'),
		(views.frontend.Preferences, '/preferences/<name>'),
		(views.frontend.Admin, '/admin'),
		(views.frontend.Admin, '/admin/<name>'),
		#(views.frontend.Post, '/post'),
		(views.redirects.About, '/about/more'),
		(views.redirects.Admin, '/about/admin'),
		(views.resources.Style, '/style-<timestamp:int>.css'),
		(views.resources.Manifest, '/manifest.json'),
		(views.resources.Media, '/media/<path:[A-z0-9.\/?]+>'),
		(views.resources.Favicon, '/favicon.ico'),
		(views.resources.Robots, '/robots.txt')
	]


	for view, route in web_routes:
		web.add_route(view.as_view(), route)


	# Media
	web.static('/static', config.frontend.join('static').str())
	#web.static('/media', config.data.join('media').str())


	# Shitpost
	web.add_route(views.redirects.HeadPats.as_view(), '/headpats')
	web.add_route(views.redirects.Socks.as_view(), '/socks')

	# Streaming websocket
	web.add_route(views.api.Streaming.as_view(), '/api/v1/streaming', stream=True)
	web.add_route(views.api.Streaming.as_view(), '/api/v1/streaming/<name>', stream=True)

	# WebUI
	#web.add_subapp('/web', webapp)
	#web.add_route('/{module}.py', modules)


def start():
	mode = 'production' if config.env == 'prod' else 'development'
	setupRoutes()

	signal.signal(signal.SIGHUP, safe_stop)
	signal.signal(signal.SIGINT, safe_stop)
	signal.signal(signal.SIGQUIT, safe_stop)
	signal.signal(signal.SIGTERM, safe_stop)

	#web.add_task(save_tokens)

	listen = config.listen
	port = config.port

	logging.info(f'Starting Barkshark Social at {listen}:{port} in {mode} mode')
	try:
		web.run(host=listen, port=port, workers=4, debug=False, access_log=False)
	except Exception as e:
		print('server exit error:', repr(e))

	safe_stop()
